=begin

Copyright 2011, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESSED OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         MakeUnique++.rb
Author:       Andreas Eisenbarth
Usage:        Select faces with the same texture,
              menu Plugins > 'Make Unique Texture ++' > 'Make selected face(s) unique'
              menu Plugins > 'Make Unique Texture ++' > 'Crop texture to selection'
Description:  The default "Make Unique texture" in SketchUp serves for
              - making a texture on a face independent from all other materials/faces
              - cropping a texture to the face
              - baking any UV distortion into the image file; thus the resulting texture pins are 
                arranged in a rectangle (especially for exporters that don't support distortion)
              This Plugin resolves several of the built-in tool's limits.
              It uses a higher quality algorithm (bicubic), crops to the best fitting bounding box
              and allows batch processing of more then one face.
Note:         If you want to get one new texture for the whole selection (instead of one for each face),
              the UV distortion can not be baked into the image (otherwise it would require unwrapping).

Version:      1.0
Date:         21.11.2011

=end
require 'sketchup.rb'



module AE

module AE::MakeUniqueTexture
@dir = File.dirname(__FILE__) # this plugin's folder
require File.join(@dir, "bounds.rb")
@temp = ( ENV['TMPDIR'] || ENV['TMP'] || ENV['TEMP'] ).gsub('\\','/')
@platform = RUBY_PLATFORM.include?('darwin')? "mac" : (RUBY_PLATFORM.include?('mswin')? (!ENV["WINEPREFIX"].nil?? "wine" : "mswin") : "other")
@unix_im_native = true
# location of portable ImageMagick
@im_win = File.expand_path(File.join(Sketchup.find_support_file("Plugins"),"ImageMagick/convert.exe"))

@max_texture_dimension = 2048

unless file_loaded?(File.basename(__FILE__))
  cmd_make_unique = UI::Command.new("Make Unique Texture"){AE::MakeUniqueTexture.make_each_unique}
  cmd_make_unique.menu_text = "Make selected face(s) unique"
  cmd_crop = UI::Command.new("Crop texture to selection"){AE::MakeUniqueTexture.crop}
  cmd_crop.menu_text = "Crop texture to selection"
#  cmd_combine_unique = UI::Command.new("Combine selected textures"){AE::MakeUniqueTexture.combine_unique}
#  cmd_combine_unique.menu_text = "Combine selected textures (unique)"
#  cmd_combine_all = UI::Command.new("Combine selected textures"){AE::MakeUniqueTexture.combine_all}
#  cmd_combine_all.menu_text = "Combine selected textures and replace all in model"
  # Menu
	menu_plugins = UI.menu("Plugins")
	menu_make_unique = menu_plugins.add_submenu("Make Unique Texture ++")
	menu_make_unique.add_item(cmd_make_unique)
	menu_make_unique.add_item(cmd_crop)
#	menu_make_unique.add_item(cmd_combine_unique)
#	menu_make_unique.add_item(cmd_combine_all)
  # Context Menu
	UI.add_context_menu_handler do |menu|
		if !Sketchup.active_model.selection.empty? && Sketchup.active_model.selection.find{|x| x.class==Sketchup::Face && (x.material&&x.material.materialType>0||x.back_material&&x.back_material.materialType>0)}
	    menu_make_unique = menu.add_submenu("Make Unique Texture ++")
	    menu_make_unique.add_item(cmd_make_unique)
	    menu_make_unique.add_item(cmd_crop)
#	    menu_make_unique.add_item(cmd_combine)
		end
  end
end



def self.make_each_unique
  return if self.imagemagick_installed? == false
  entities = Sketchup.active_model.selection.to_a
  return UI.messagebox("Select textured faces that you want to make unique.") if entities.empty?
  Sketchup.active_model.start_operation("Make unique texture")
  entities.each{|e|
    next if e.class != Sketchup::Face
    self.make_unique(e, true) if e.material
    self.make_unique(e, false) if e.back_material
  }
  Sketchup.active_model.commit_operation
end #def each_single_face



##
# Replace the texture on one face with a new, unique texture
#
def self.make_unique(face, front=true)
  return puts("AE::MakeUniqueTexture: no face given") if face.nil? || face.class != Sketchup::Face
  return puts("AE::MakeUniqueTexture: face has no texture") if (front && !face.material || !front && !face.back_material) || (front && face.material && face.material.materialType==0 || !front && face.back_material && face.back_material.materialType==0)
  @model = Sketchup.active_model
  @tw = Sketchup.create_texture_writer
  begin
    Sketchup.status_text = "Make unique texture: Processing..."
    m = (front)? face.material : face.back_material
    # export texture
    texturepath = self.export_texture(m)
    return if texturepath.nil?
    # get optimal bounding box
    b = AE::Bounds.new(face)
    points = [b.corner(2), b.corner(0), b.corner(1), b.corner(3)] # counter-clockwise from top left
    # get UV coordinates
    uv_help = face.get_UVHelper(true, true, @tw)
    uvqs = points.collect{|p| (front)? uv_help.get_front_UVQ(p) : uv_help.get_back_UVQ(p)}
    uvqs.each{|uv| uv.x/=uv.z; uv.y/=uv.z; uv.z=1}
    # get texture resolution
    r = self.texture_resolution(face, front)
    # calculate new texture size
    w = m.texture.image_width
    h = m.texture.image_height
    new_w = (r * b.width).round.to_i
    new_h = (r * b.depth).round.to_i
    new_w = (w > @max_texture_dimension)? w : @max_texture_dimension if new_w > @max_texture_dimension    
    new_h = (h > @max_texture_dimension)? h : @max_texture_dimension if new_h > @max_texture_dimension
    # create command for ImageMagick
    # convert image.jpg -virtual-pixel tile -distort Perspective '0,0 0,0  0,500 0,250  500,500 500,500  500,0 500,0' image.jpg
    xy = []
    uvqs.each{|uv| xy << (uv[0]*w).to_i.to_s+","+(h-(uv[1]*h)).to_i.to_s}
    convert_option = %[-virtual-pixel tile -distort Perspective ]
    convert_option +=  %["#{xy[0]} 0,0  #{xy[1]} 0,#{h} ]
    convert_option += %[ #{xy[2]} #{w},#{h}  #{xy[3]} #{w},0"]
    convert_option += %[ -resize #{new_w}x#{new_h}!]
    self.run_imagemagick(texturepath, convert_option){
      # import new texture
      new_material = self.import_texture(texturepath, m)
      new_material.texture.size = [b.width, b.height]
      # position texture
      pt_array = [points[0], [0,1,1], points[1], [0,0,1], points[2], [1,0,1], points[3], [1,1,1]]
      face.position_material(new_material, pt_array, front)
      File.delete(texturepath) if File.exists?(texturepath)
    }
  rescue Exception => e
    File.delete(texturepath) if texturepath && File.exists?(texturepath)
    @model.abort_operation
    puts("Error #<#{e.class.name}:#{e.message}.>")
    puts(e.backtrace) if $VERBOSE
  ensure
    Sketchup.status_text = ""
  end #begin/rescue
end #def process



##
# Replace the texture on all face with one common, unique texture
# This method does not distort the texture; it applies the BoundingBox in UV space.
#
def self.crop(entities=Sketchup.active_model.selection.to_a, front=true)
  @model = Sketchup.active_model
  return if self.imagemagick_installed? == false
  faces = entities.find_all{|e| e.class == Sketchup::Face}
  return UI.messagebox("Select textured faces with the same material that you want to make unique.") if faces.empty?
  i = 0; (m = (front)? faces[i].material : faces[i].back_material; i+=1) while i < faces.length && (m==nil || m.materialType==0)
  faces = faces.find_all{|f| fm = (front)? f.material : f.back_material; fm != nil && fm == m}
  return UI.messagebox("Selected faces have no textured front material.") if faces.empty?
  begin
    @model.start_operation("Crop texture to selection")
    Sketchup.status_text = "Crop texture to selection: Processing..."
    @tw = Sketchup.create_texture_writer
    all_points = [] # all points to be considered for the BoundingBox
    # collect UV positions of other faces
    @faces = {}
    faces.each{|f|
      ps = f.vertices.collect{|v| v.position}
      uv_help = f.get_UVHelper(true, true, @tw)
      # translate all UVs so that the center of the face is in the range 0-1
      # this is necessary if faces share a repeated tile of the same area on the texture
      cuv = (front)? uv_help.get_front_UVQ(f.bounds.center) : uv_help.get_back_UVQ(f.bounds.center)
      cuv.x/=cuv.z; cuv.y/=cuv.z; cuv.z=1
      tr = [-cuv[0].divmod(1)[0], -cuv[1].divmod(1)[0], 0]
      # convert positions into uvq, then from mainface back into xyz
      @faces[f] = ps.collect{|p|
        uv = (front)? uv_help.get_front_UVQ(p) : uv_help.get_back_UVQ(p)
        uv.x/=uv.z; uv.y/=uv.z; uv.z=1
        uv += tr
        all_points << uv
        uv
      }
      if ps.length == 3 # triangle: add a fourth point for better uvq, but don't consider it for the BoundingBox
        p_triangle = ps[0] + ps[1].vector_to(ps[2])
        uv = (front)? uv_help.get_front_UVQ(p_triangle) : uv_help.get_back_UVQ(p_triangle)
        uv.x/=uv.z; uv.y/=uv.z; uv.z=1
        uv += tr
        @faces[f] << uv
      end
    }
    # export texture
    texturepath = self.export_texture(m)
    return if texturepath.nil?
    # get optimal bounding box for all points
    b = AE::Bounds.new(all_points) # this time in UV space
    new_uvs = [b.corner(2), b.corner(0), b.corner(1), b.corner(3)] # counter-clockwise from top left
    # get texture resolution
    w = m.texture.image_width
    h = m.texture.image_height
    r = Math.sqrt(w**2 + h**2)/(b.diagonal)
    # calculate new texture size
    new_w = (r * b.width).to_i
    new_h = (r * b.depth).to_i
    # create command for ImageMagick
    xy = []
    new_uvs.each{|uv| xy << (uv[0]*w).to_i.to_s+","+(h-(uv[1]*h)).to_i.to_s}
    convert_option = %[-virtual-pixel tile -distort Perspective ]
    convert_option +=  %["#{xy[0]} 0,0  #{xy[1]} 0,#{h} ]
    convert_option += %[ #{xy[2]} #{w},#{h}  #{xy[3]} #{w},0"]
    convert_option += %[ -resize #{new_w}x#{new_h}!]
    self.run_imagemagick(texturepath, convert_option){
      # import new texture
      new_m = self.import_texture(texturepath, m)
      new_m.texture.size = [b.width, b.height]
      # position texture on all faces
      t = Geom::Transformation.new(new_uvs[1], new_uvs[1].vector_to(new_uvs[2]), new_uvs[2].vector_to(new_uvs[3]))
      t = t * Geom::Transformation.scaling(new_uvs[1].distance(new_uvs[2]), new_uvs[2].distance(new_uvs[3]), 1)
      @faces.each{|f, uvs|
        ps = f.vertices.collect{|v| v.position}
        ps << ps[0] + ps[1].vector_to(ps[2]) if ps.length == 3 # triangle
        uvqs = uvs.collect{|uv| uv.transform(t.inverse) }
        pt_array = []
        4.times{|i| pt_array << ps[i]; pt_array << uvqs[i]}
        f.position_material(new_m, pt_array, front)
      }
      File.delete(texturepath) if File.exists?(texturepath)
      @model.commit_operation
    }
  rescue Exception => e
    File.delete(texturepath) if texturepath && File.exists?(texturepath)
    @model.abort_operation
    puts("Error #<#{e.class.name}:#{e.message}.>")
    puts(e.backtrace) if $VERBOSE
  ensure
    Sketchup.status_text = ""
  end #begin/rescue
end



##
# export a texture to texturepath
#
def self.export_texture(material)
  texturepath = self.to_filename(@temp, File.basename(material.texture.filename))
  # create temporary group to export texture image without uv-mapping
  tmp_group = @model.entities.add_group
  tmp_group.material = material
  @tw.load(tmp_group)
  success = @tw.write(tmp_group, texturepath, true)
  puts(material.name+": could not write file "+texturepath) if success!=0
  @model.entities.erase_entities(tmp_group)
  return (success==0)? texturepath : nil
end #def import_texture



##
# import a texture from texturepath, inherit properties from similar material
#
def self.import_texture(texturepath, similar_m=nil)
  return (puts("file "+texturepath+" not found"); similar_m) if !File.exists?(texturepath)
  i = 0; new_name = (similar_m)? similar_m.display_name : File.basename(texturepath, File.extname(texturepath))
  (new_name = similar_m.display_name + i.to_s; i+=1) while @model.materials[new_name]
  m = @model.materials.add(new_name)
  m.texture = texturepath
  if similar_m
    m.texture.size = [similar_m.texture.width, similar_m.texture.height]
    #m.color = similar_m.color
    m.alpha = similar_m.alpha
  end
  return m
end #def import_texture



##
# get texture resolution of a face
#
def self.texture_resolution(face, front=true)
  m = (front)? face.material : face.back_material
  # get some test points in object space (here: 8)
  c = face.bounds.center
  vec0 = c.vector_to(face.vertices[1].position).normalize
  t = Geom::Transformation.new(c, face.normal, 45.degrees)
  ps = [c+vec0]; (2..8).each{|i| ps << ps.last.transform(t)}
  # in texture space
  uv_help = face.get_UVHelper(true, true, @tw)
  uvs = ps.collect{|p| (front)? uv_help.get_front_UVQ(p) : uv_help.get_back_UVQ(p) }
  uvs.collect!{|uv| uv.to_a.collect!{|q| q /= uv[2]} }
  # texture space lengths corresponding to object space length 1
  uvlengths = (0..uvs.length-1).collect{|i| uvs[i].distance(uvs[i-4])/2.0 }
  w = m.texture.image_width
  h = m.texture.image_height
  # one model unit (inch) in texture units (pixels)
  r = Math.sqrt(w*h) * uvlengths.max # in direction of highest resolution; alternative: arithmetic or geometric mean
  r *= 1.5 if uvlengths.max-uvlengths.min > 0.001 # increase resolution slightly if texture is distorted
  return r
end #def texture_resolution



##
# Check if ImageMagick is installed, otherwise offer selection of alternative install location
#
def self.imagemagick_installed?
  @im_win = Sketchup.read_default("ImageMagick", "location", @im_win) if !File.exists?(@im_win)
  if !File.exists?(@im_win)
    UI.messagebox("Make Unique Texture HQ requires ImageMagick, but it could not be found. \nPlease navigate to the ImageMagick folder and select the file 'convert.exe' or install ImageMagick from \nhttp://www.imagemagick.org/script/binary-releases.php.",MB_OK)
    path = UI.openpanel("Please select the file path of convert.exe", @im_win, "convert.exe")
    Sketchup.write_default("ImageMagick", "location", path)
    return false if !File.exists?(path.to_s)
  end
end #def imagemagick_installed?



##
# Execute an ImageMagick command
#
def self.run_imagemagick(texturepath, convert_option, &block)
  if @platform == "wine" && File.exists?("Z:/usr/bin/convert") && @unix_im_native == true
    executable = 'convert'
    texturepath = %["`winepath -u '#{texturepath}'`"]
  elsif @platform == "mswin" || @platform == "wine"
    executable = '"' + @im_win.gsub(/\//,'//').gsub(/\//,'\\') + '"'
    texturepath = '"' + texturepath.gsub(/\//,'//').gsub(/\//,'\\') + '"'
  elsif @platform == "mac"
    executable = 'convert'
  end
  cmd = executable + ' ' + texturepath + ' ' + convert_option + ' ' + texturepath
  # execute shell command
  self.run_shell_command(cmd, &block)
end



##
# Execute a command / a list of commands <cmd>
#
# modified, based on Fredo6's script http://forums.sketchucation.com/viewtopic.php?f=180&t=36973
# - on Windows: launches a vbs script to avoid a black console window
# - launches a batch file asynchronously in order not to run in SketchUp's process
# - continues optionally with a callback, after the batch script is finished
#
def self.run_shell_command(cmd, &block)
  cmd = [cmd] if cmd.class != Array
  finished = File.join(@temp, "finished.txt")
  File.delete(finished) if File.exists?(finished)

  # Windows
  if @platform == "mswin"
    # create a batch script with the commands
    fbat = File.join(@temp, "temporary.bat")
    File.open(fbat, "w"){|f|
      f.puts "@echo off"
      cmd.each{|c| f.puts(c) }
      f.puts %[echo "finished" > "#{@temp.gsub(/\//,'//').gsub(/\//,'\\')}\\finished.txt"]
    }
    # create a Visual Basic file to launch the batch script without black console window
    fvbs = File.join(@temp, "temporary.vbs")
    File.open(fvbs, "w"){|f|
      f.puts %[Set WshShell = CreateObject("WScript.Shell")]
      f.puts %[WshShell.Run """#{fbat}""", 0]
      # a third argument bWaitOnReturn=true would wait for the script to finish (synchronously),
      # but uses SketchUp's process and is 20 times slower.
      # f.puts %[WshShell.Run """#{fbat}""", 0, true]
      f.puts %[WScript.Quit]
    }
    system(%[wscript #{fvbs}])

  # Mac OS X
  elsif @platform == "mac"
    fsh = File.join(@temp, "temporary.sh")
    File.open(fsh, "w"){|f|
      f.puts %[ #!/bin/sh]
      cmd.each{|c| f.print(c+";") }
      f.puts %[touch "#{@temp}/finished.txt";]
    }
    system(%[bash #{fsh}])

  # other
  elsif @platform == "wine"
    # check for native ImageMagick installation
    if File.exists?("Z:/bin/bash") && File.exists?("Z:/usr/bin/convert") && @unix_im_native == true
      fsh = File.join(@temp, "temporary.sh")
      File.open(fsh, "w"){|f|
        f.puts %[ #!/bin/sh]
        cmd.each{|c| f.print(c+";") }
        f.puts %[touch "`winepath -u '#{@temp}'`/finished.txt";]
      }
      begin
        system(%[Z:\\bin\\bash "#{ENV["WINEPREFIX"]}/dosdevices/#{fsh.sub(/^(\w)(?=\:)/,@dir[0,1].downcase)}"])
      rescue
      end
    # otherwise use portable Windows ImageMagick as distributed with this plugin
    #   no WScript necessary because Wine does not show black console window
    else
      # create a batch script with the commands
      fbat = File.join(@temp, "temporary.bat")
      File.open(fbat, "w"){|f|
        f.puts "@echo off"
        cmd.each{|c| f.puts(c) }
        f.puts %[echo "finished" > "#{@temp.gsub(/\//,'//').gsub(/\//,'\\')}\\finished.txt"]
      }
      system(%[#{fbat}])
    end
  end
  # since the batch script runs asynchronously, we need to know when it is finished
  # the batch script creates a "finished.txt" file, and we continue only when it is created
  self.file_observer(finished){
    [finished,fbat,fvbs,fsh].each{|f| File.delete(f) if f && File.exists?(f)}
    block.call
  }
end #def run_shell_command



##
# observe when a file exists and execute block
#
def self.file_observer(file, &block)
  if !File.exists?(file)
    UI.start_timer(0.1){ self.file_observer(file, &block) }
  else
    File.delete(file)
    block.call
  end
end #def file_observer



##
# Convert a random string into a proper, unique filename
# arguments:
#   dir: directory where the file should go
#   string: any user input
#   ext: force a specific extension if the string doesn't contain one
# returns: a complete file path
# 
def self.to_filename(dir, string, ext=nil)
  # separate extension from basename
  ext = string[/\.\w{1,3}$/].to_s if ext.nil?
  string.sub!(/#{ext}$/,"")
  # clean up basename
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.\(\)\#]+/,"_").gsub(/^_+|_+$/,"").to_s
  string = "file" if string.empty?
  base = File.join(dir, string)
  # detect collision of filenames and return alternative filename (numbered)
  base_orig = base
  i = 0
  while File.exists?(base + ext)
    base = base_orig + i.to_s
    i += 1
  end
  return base + ext
end #def to_filename



end #module AE::MakeUniqueTexture



end #module AE



file_loaded(File.basename(__FILE__))
